/* 
 * $TSUKUBA_Release: Omni OpenMP/XcalableMP Compiler Software Version 0.5.1 $
 * $TSUKUBA_Copyright:
 *  Copyright (C) 2010 University of Tsukuba
 *  
 *  This software is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License version
 *  2.1 published by the Free Software Foundation.
 *  
 *  Please check the Copyright and License information in the files named
 *  COPYRIGHT and LICENSE under the top  directory of the Omni Compiler
 *  Software release kit.
 *  
 *  * The specification of XcalableMP has been designed by the XcalableMP
 *    Specification Working Group (http://www.xcalablemp.org/).
 *  
 *  * The development of this software is supported by "Seamless and
 *    Highly-productive Parallel Programming Environment for
 *    High-performance computing" project funded by Ministry of Education,
 *    Culture, Sports, Science and Technology, Japan.
 *  $
 */
package xcodeml.c.decompile;

import xcodeml.XmException;
import xcodeml.c.obj.XcNode;
import xcodeml.c.util.XmcWriter;

/**
 * Internal object represents coArrayAssignExpr.
 */
public class XcXmpCoArrayAssignObj extends XcObj implements XcExprObj
{
    XcExprObj _dst;

    XcExprObj _src;

    @Override
    public void appendCode(XmcWriter w) throws XmException
    {
        XcExprObj obj = XcXmpFactory.createCoAPutFuncCall(this);

        w.add(obj);
    }

    /**
     * Gets expression object of coaray put destination.
     * 
     * @return a coarray assignment destination.
     */
    public XcExprObj getDst()
    {
        return _dst;
    }

    /**
     * Gets expression object of coaray put source.
     * 
     * @return a coarray assignment source.
     */
    public XcExprObj getSrc()
    {
        return _src;
    }

    @Override
    public void addChild(XcNode child)
    {
        if(child instanceof XcExprObj) {
            if(_dst == null)
                _dst = (XcExprObj)child;
            else
                _src = (XcExprObj)child;
        } else {
            throw new IllegalArgumentException(child.getClass().getName());
        }
    }

    @Override
    public void checkChild()
    {
        if(_src == null || _dst == null)
            throw new IllegalArgumentException();
    }

    @Override
    public XcNode[] getChild()
    {
        return toNodeArray(_dst, _src);
    }

    @Override
    public void setChild(int index, XcNode child)
    {
        if(index == 0) {
            if(child instanceof XcXmpCoArrayRefObj) {
                _dst = (XcXmpCoArrayRefObj)child;
                return;
            } else
                throw new IllegalArgumentException(child.getClass().getName());
        }

        if(index == 1) {
            if(child instanceof XcExprObj) {
                _src = (XcExprObj)child;
                return;
            } else
                throw new IllegalArgumentException(child.getClass().getName());
        }

        throw new IllegalArgumentException(index + ":" + child.getClass().getName());
    }
}